# *** imports

# ** core
from typing import Dict, Any

# ** app
from .feature import FeatureContext
from .error import ErrorContext
from ..configs.app import DEFAULT_ATTRIBUTES
from ..models.feature import Request
from ..models.app import *
from ..handlers.app import (
    AppService,
    AppHandler,
    AppRepository,
)
from ..commands import (
    Command,
    TiferetError
)
from ..commands.app import GetAppInterface


# *** contexts

# ** context: app_context
class AppContext(object):
    '''
    The application context is a class that is used to create and run the application instance.
    '''

    # * attribute: settings
    settings: Dict[str, Any]

    # * attribute: app_service
    app_service: AppService

    # * method: init
    def __init__(self, settings: Dict[str, Any] = {}, app_service: AppService = AppHandler()):
        '''
        Initialize the application context.
        
        :param settings: The application settings.
        :type settings: dict
        :param app_service: The application service to use for executing app requests.
        :type app_service: AppService
        '''
        
        # Assign the settings and app service.
        self.settings = settings
        self.app_service = app_service

    # * method: load_interface
    def load_interface(self, interface_id: str) -> 'AppInterfaceContext':
        '''
        Load the application interface.

        :param interface_id: The interface ID.
        :type interface_id: str
        :return: The application interface context.
        :rtype: AppInterfaceContext
        '''

        # Load the app repository.
        app_repo: AppRepository = self.app_service.load_app_repository(**self.settings)

        # Get the app interface settings.
        app_interface = Command.handle(
            GetAppInterface,
            dependencies=dict(
                app_repo=app_repo
            ),
            interface_id=interface_id
        )

        # Retrieve the default attributes from the configuration.
        default_attrs = [ModelObject.new(
            AppAttribute,
            **attr_data,
            validate=False
        ) for attr_data in DEFAULT_ATTRIBUTES]
        
        # Create the app interface context.
        app_interface_context = self.app_service.load_app_instance(app_interface, default_attrs=default_attrs)

        # Verify that the app interface context is valid.
        if not isinstance(app_interface_context, AppInterfaceContext):
            raise TiferetError(
                'APP_INTERFACE_INVALID',
                f'App context for interface is not valid: {interface_id}.',
                interface_id
            )
        
        # Return the app interface context.
        return app_interface_context

    # * method: run
    def run(self,
            interface_id: str,
            feature_id: str,
            headers: Dict[str, str] = {},
            data: Dict[str, Any] = {},
            debug: bool = False,
            **kwargs
        ) -> Any:
        '''
        Run the application interface.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param dependencies: The dependencies.
        :type dependencies: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The response.
        :rtype: Any
        '''

        # Load the interface.
        app_interface = self.load_interface(interface_id)

        # Run the interface.
        return app_interface.run(
            feature_id, 
            headers, 
            data, 
            debug=debug,
            **kwargs
        )


# ** context: app_interface_context
class AppInterfaceContext(object): 
    '''
    The application interface context is a class that is used to create and run the application interface.
    '''

    # * attribute: interface_id
    interface_id: str

    # * attribute: features
    features: FeatureContext

    # * attribute: errors
    errors: ErrorContext

    # * method: init
    def __init__(self, interface_id: str, features: FeatureContext, errors: ErrorContext):
        '''
        Initialize the application interface context.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param features: The feature context.
        :type features: FeatureContext
        :param errors: The error context.
        :type errors: ErrorContext
        '''

        # Assign instance variables.
        self.interface_id = interface_id
        self.features = features
        self.errors = errors

    # * method: parse_request
    def parse_request(self, headers: Dict[str, str] = {}, data: Dict[str, Any] = {}) -> Request:
        '''
        Parse the incoming request.

        :param headers: The request headers.
        :type headers: dict
        :param data: The request data.
        :type data: dict
        :return: The parsed request.
        :rtype: Request
        '''

        # Add the interface id to the request headers.
        headers.update(dict(
            interface_id=self.interface_id,
        ))

        # Create the request model object.
        request = ModelObject.new(
            Request,
            headers=headers,
            data=data,
        )

        # Return the request model object.
        return request
    
    # * method: execute_feature
    def execute_feature(self, feature_id: str, request: Request, **kwargs):
        '''
        Execute the feature context.

        :param feature_id: The feature identifier.
        :type feature_id: str
        :param request: The request.
        :type request: Request
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Add the feature id to the request headers.
        request.headers.update(dict(
            feature_id=feature_id
        ))

        # Execute feature context and return session.
        self.features.execute_feature(feature_id, request, **kwargs)

    # * method: handle_error
    def handle_error(self, error: Exception) -> Any:
        '''
        Handle the error and return the response.

        :param error: The error to handle.
        :type error: Exception
        :return: The error response.
        :rtype: Any
        '''

        # If the error is not a TiferetError, wrap it in one.
        if not isinstance(error, TiferetError):
            error = TiferetError(
                'APP_ERROR',
                f'An error occurred in the app: {str(error)}',
                str(error)
            )

        # Handle the error and return the response.
        return self.errors.handle_error(error)

    # * method: handle_response
    def handle_response(self, request: Request) -> Any:
        '''
        Handle the response from the request.

        :param request: The request context.
        :type request: RequestContext
        :return: The response.
        :rtype: Any
        '''

        # Handle the response and return it.
        return request.handle_response()
    
    # * method: run
    def run(self, 
            feature_id: str, 
            headers: Dict[str, str] = {}, 
            data: Dict[str, Any] = {},
            **kwargs) -> Any:
        '''
        Run the application interface by executing the feature.

        :param feature_id: The feature identifier.
        :type feature_id: str
        :param headers: The request headers.
        :type headers: dict
        :param data: The request data.
        :type data: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''
        
        # Parse request.
        request = self.parse_request(headers, data)

        # Execute feature context and return session.
        try:
            self.execute_feature(feature_id, request, **kwargs)

        # Handle error and return response if triggered.
        except TiferetError as e:
            return self.handle_error(e)

        # Handle response.
        return self.handle_response(request)



# *** imports

# ** core
from typing import Any, List, Tuple, Dict

# ** app
from .core import *
from ...data.container import ContainerAttributeYamlData
from ...contracts.container import ContainerRepository, ContainerAttribute


# *** proxies

# ** proxy: container_yaml_proxy
class ContainerYamlProxy(ContainerRepository, YamlConfigurationProxy):
    '''
    Yaml proxy for container attributes.
    '''

    # * init
    def __init__(self, container_config_file: str):
        '''
        Initialize the yaml proxy.
        
        :param container_config_file: The YAML file path for the container configuration.
        :type container_config_file: str
        '''

        # Set the container configuration file.
        super().__init__(container_config_file)

    # * method: load_yaml
    def load_yaml(self, start_node: callable = lambda data: data, create_data: callable = lambda data: data) -> Any:
        '''
        Load data from the YAML configuration file.
        :param start_node: The starting node in the YAML file.
        :type start_node: str
        :param create_data: A callable to create data objects from the loaded data.
        :type create_data: callable
        :return: The loaded data.
        :rtype: Any
        '''

        # Load the YAML file contents using the yaml config proxy.
        try:
            return super().load_yaml(
                start_node=start_node,
                create_data=create_data
            )
        
        # Raise an error if the loading fails.
        except (Exception, TiferetError) as e:
            raise_error.execute(
                'CONTAINER_CONFIG_LOADING_FAILED',
                f'Unable to load container configuration file {self.config_file}: {e}.',
                self.config_file,
                str(e)
            )

    # * method: get_attribute
    def get_attribute(self, attribute_id: str) -> ContainerAttribute:
        '''
        Get the attribute from the yaml file.

        :param attribute_id: The attribute id.
        :type attribute_id: str
        :return: The container attribute.
        :rtype: ContainerAttribute
        '''

        # Load the attribute data from the yaml configuration file.
        data = self.load_yaml(
            create_data=lambda data: ContainerAttributeYamlData.from_data(
                id=attribute_id, **data),
            start_node=lambda data: data.get('attrs').get(attribute_id),
        )

        # If the data is None or the type does not match, return None.
        # Remove the type logic later, as the type parameter will be removed in v2 (obsolete).
        if data is None:
            return None
        
        # Return the attribute.
        return data.map()

    # * method: list_all
    def list_all(self) -> Tuple[List[ContainerAttribute], Dict[str, str]]:
        '''
        List all the container attributes and constants.

        :return: The list of container attributes and constants.
        :rtype: Tuple[List[ContainerAttribute], Dict[str, str]]
        '''

        # Define create data function to parse the YAML file.
        def create_data(data):
            
            # Create a list of ContainerAttributeYamlData objects from the YAML data.
            attrs = [
                ContainerAttributeYamlData.from_data(id=id, **attr_data)
                for id, attr_data in data.get('attrs', {}).items()
            ] if data.get('attrs') else []

            # Get the constants from the YAML data.
            consts = data.get('const', {}) if data.get('const') else {}

            # Return the parsed attributes and constants.
            return attrs, consts

        # Load the attribute data from the yaml configuration file.
        attr_data, consts = self.load_yaml(
            create_data=create_data
        )

        # Return the list of container attributes.
        return (
            [data.map() for data in attr_data],
            consts
        )



# *** imports

# ** core
from typing import Any

# ** app
from .core import *
from ...data import DataObject
from ...data.app import AppInterfaceYamlData
from ...contracts.app import AppRepository, AppInterface


# *** proxies

# ** proxy: app_yaml_proxy
class AppYamlProxy(AppRepository, YamlConfigurationProxy):


    # * method: init
    def __init__(self, app_config_file: str):
        '''
        Initialize the YAML proxy.

        :param app_config_file: The application configuration file.
        :type app_config_file: str
        '''

        # Set the configuration file.
        super().__init__(app_config_file)

    # * method: load_yaml
    def load_yaml(self, start_node: callable = lambda data: data, create_data: callable = lambda data: data) -> Any:
        '''
        Load data from the YAML configuration file.
        :param start_node: The starting node in the YAML file.
        :type start_node: str
        :param create_data: A callable to create data objects from the loaded data.
        :type create_data: callable
        :return: The loaded data.
        :rtype: Any
        '''

        # Load the YAML file contents using the yaml config proxy.
        try:
            return super().load_yaml(
                start_node=start_node,
                create_data=create_data
            )
        
        # Raise an error if the loading fails.
        except (Exception, TiferetError) as e:
            raise_error.execute(
                'APP_CONFIG_LOADING_FAILED',
                f'Unable to load app configuration file {self.config_file}: {e}.',
                self.config_file,
                str(e)
            )

    # * method: list_interfaces
    def list_interfaces(self) -> list[AppInterface]:
        '''
        List all app interfaces.

        :return: The list of app interfaces.
        :rtype: List[AppInterface]
        '''

        # Load the app interface data from the yaml configuration file and map it to the app interface object.
        interfaces = self.load_yaml(
            create_data=lambda data: [
                DataObject.from_data(
                    AppInterfaceYamlData,
                    id=interface_id,
                    **record
                ).map() for interface_id, record in data.items()],
            start_node=lambda data: data.get('interfaces'))

        # Return the list of app interface objects.
        return interfaces

    # * method: get_interface
    def get_interface(self, id: str) -> AppInterface:
        '''
        Get the app interface.

        :param id: The app interface id.
        :type id: str
        :return: The app interface.
        :rtype: AppInterface
        '''

        # Load the app interface data from the yaml configuration file.
        _data: AppInterface = self.load_yaml(
            create_data=lambda data: DataObject.from_data(
                AppInterfaceYamlData,
                id=id, 
                **data
            ),
            start_node=lambda data: data.get('interfaces').get(id)
        )

        # Return the app interface object.
        # If the data is None, return None.
        try:
            return _data.map()
        except AttributeError:
            return None


# *** imports

# ** core
from typing import Any

# ** app
from ...commands import *
from ...clients import yaml_client


# *** classes

# ** class yaml_proxy
class YamlConfigurationProxy(object):
    '''
    A base class for proxies that handle YAML configuration files.
    '''

    # * field: config_file
    config_file: str = None

    # * method: init
    def __init__(self, config_file: str):
        '''
        Initialize the proxy.

        :param config_file: The configuration file.
        :type config_file: str
        '''
        self.config_file = config_file

    # * method: load_yaml
    def load_yaml(self, start_node: callable = lambda data: data, create_data: callable = lambda data: data) -> Any:
        '''
        Load data from the YAML configuration file.

        :param start_node: The starting node in the YAML file.
        :type start_node: str
        :param create_data: A callable to create data objects from the loaded data.
        :type create_data: callable
        :return: The loaded data.
        :rtype: any
        '''

        # Load the YAML file using the yaml client.
        try:
            return yaml_client.load(
                self.config_file,
                create_data=create_data,
                start_node=start_node
            )
        except FileNotFoundError as e:
            # If the file is not found, raise an error.
            raise_error.execute(
                'CONFIG_FILE_NOT_FOUND',
                f'Configuration file {self.config_file} not found.',
                self.config_file
            )


# *** imports


# *** configs

# ** config: errors
ERRORS = [
    dict(
        id='parameter_parsing_failed',
        name='Parameter Parsing Failed',
        error_code='PARAMETER_PARSING_FAILED',
        message=[
            dict(lang='en_US', text='Failed to parse parameter: {}. Error: {}')
        ]
    ),
    dict(
        id='import_dependency_failed',
        name='Import Dependency Failed',
        error_code='IMPORT_DEPENDENCY_FAILED',
        message=[
            dict(lang='en_US', text='Failed to import dependency: {} from module {}. Error: {}')
        ]
    ),
    dict(
        id='invalid_dependency_error',
        name='Invalid Dependency Error',
        error_code='INVALID_DEPENDENCY_ERROR',
        message=[
            dict(lang='en_US', text='Dependency {} could not be resolved: {}')
        ]
    ),
    dict(
        id='app_repository_import_failed',
        name='App Repository Import Failed',
        error_code='APP_REPOSITORY_IMPORT_FAILED',
        message=[
            dict(lang='en_US', text='Failed to import app repository: {}.')
        ]
    ),
    dict(
        id='app_interface_not_found',
        name='App Interface Not Found',
        error_code='APP_INTERFACE_NOT_FOUND',
        message=[
            dict(lang='en_US', text='App interface with ID {} not found.')
        ]
    ),
    dict(
        id='feature_command_loading_failed',
        name='Feature Command Loading Failed',
        error_code='FEATURE_COMMAND_LOADING_FAILED',
        message=[
            dict(lang='en_US', text='Failed to load feature command attribute: {}. Ensure the container attributes are configured with the appropriate default settings/flags. {}')
        ]
    ),
    dict(
        id='error_not_found',
        name='Error Not Found',
        error_code='ERROR_NOT_FOUND',
        message=[
            dict(lang='en_US', text='Error not found: {}.')
        ]
    ),
    dict(
        id='container_attributes_not_found',
        name='Container Attributes Not Found',
        error_code='CONTAINER_ATTRIBUTES_NOT_FOUND',
        message=[
            dict(lang='en_US', text='No container attributes provided to load the container.')
        ]
    ),
    dict(
        id='dependency_type_not_found',
        name='Dependency Type Not Found',
        error_code='DEPENDENCY_TYPE_NOT_FOUND',
        message=[
            dict(lang='en_US', text='No dependency type found for attribute {} with flags {}.')
        ]
    ),
    dict(
        id='request_not_found',
        name='Request Not Found',
        error_code='REQUEST_NOT_FOUND',
        message=[
            dict(lang='en_US', text='Request data is not available for parameter parsing.')
        ]
    ),
    dict(
        id='parameter_not_found',
        name='Parameter Not Found',
        error_code='PARAMETER_NOT_FOUND',
        message=[
            dict(lang='en_US', text='Parameter {} not found in request data.')
        ]
    ),
    dict(
        id='feature_not_found',
        name='Feature Not Found',
        error_code='FEATURE_NOT_FOUND',
        message=[
            dict(lang='en_US', text='Feature with ID {} not found.')
        ]
    ),
    dict(
        id='app_error',
        name='App Error',
        error_code='APP_ERROR',
        message=[
            dict(lang='en_US', text='An error occurred in the app: {}')
        ]
    ),
    dict(
        id='config_file_not_found',
        name='Configuration File Not Found',
        error_code='CONFIG_FILE_NOT_FOUND',
        message=[
            dict(lang='en_US', text='Configuration file {} not found.')
        ]
    ),
    dict(
        id='app_config_loading_failed',
        name='App Configuration Loading Failed',
        error_code='APP_CONFIG_LOADING_FAILED',
        message=[
            dict(lang='en_US', text='Unable to load app configuration file {}: {}.')
        ]
    ),
    dict(
        id='container_config_loading_failed',
        name='Container Configuration Loading Failed',
        error_code='CONTAINER_CONFIG_LOADING_FAILED',
        message=[
            dict(lang='en_US', text='Unable to load container configuration file {}: {}.')
        ]
    ),
    dict(
        id='feature_config_loading_failed',
        name='Feature Configuration Loading Failed',
        error_code='FEATURE_CONFIG_LOADING_FAILED',
        message=[
            dict(lang='en_US', text='Unable to load feature configuration file {}: {}.')
        ]
    ),
    dict(
        id='error_config_loading_failed',
        name='Error Configuration Loading Failed',
        error_code='ERROR_CONFIG_LOADING_FAILED',
        message=[
            dict(lang='en_US', text='Unable to load error configuration file {}: {}.')
        ]
    )
]


# *** configs

# ** config: default_attributes
DEFAULT_ATTRIBUTES = [
    dict(
        attribute_id='container_repo',
        module_path='tiferet.proxies.yaml.container',
        class_name='ContainerYamlProxy',
        parameters=dict(
            container_config_file='app/configs/container.yml'
        )
    ),
    dict(
        attribute_id='feature_repo',
        module_path='tiferet.proxies.yaml.feature',
        class_name='FeatureYamlProxy',
        parameters=dict(
            feature_config_file='app/configs/feature.yml'
        )
    ),
    dict(
        attribute_id='error_repo',
        module_path='tiferet.proxies.yaml.error',
        class_name='ErrorYamlProxy',
        parameters=dict(
            error_config_file='app/configs/error.yml'
        )
    ),
    dict(
        attribute_id='container_service',
        module_path='tiferet.handlers.container',
        class_name='ContainerHandler',
    ),
    dict(
        attribute_id='feature_service',
        module_path='tiferet.handlers.feature',
        class_name='FeatureHandler',
    ),
    dict(
        attribute_id='error_service',
        module_path='tiferet.handlers.error',
        class_name='ErrorHandler',
    ),
    dict(
        attribute_id='container',
        module_path='tiferet.contexts.container',
        class_name='ContainerContext',
    ),
    dict(
        attribute_id='features',
        module_path='tiferet.contexts.feature',
        class_name='FeatureContext',
    ),
    dict(
        attribute_id='errors',
        module_path='tiferet.contexts.error',
        class_name='ErrorContext',
    ),
]